#include <stdio.h>
#include <math.h>
#include <limits.h>
#include "war.h"
#include "hash.h"

double	OfficerA = 0.2;
double	OfficerB = 70.;
double	FightingBias = 0.5;
unsigned long	Step = 0;
unsigned long	nCountries = 0;
unsigned long	nFights = 0, nOccupation = 0,
		nPopulation = 0, nWarDead = 0;
Country *theCountries = NULL;
static Country *freeC = NULL;
static unsigned long countryID = 0;
unsigned long TracedCountryID = 0;
LogData	logPopulation = {3, 0., 0., NULL, 0, 0},
	logTerritory = {1, 0., 0., NULL, 0, 0},
	logNCountries = {1, 0., 0., NULL, 0, 0},
	logMaxNVillages = {1, 0., 0., NULL, 0, 0},
	logAveNVillages = {1, 0., 0., NULL, 0, 0},
	logNFights = {2, 0., 0., NULL, 0, 0},
	logWarDead = {1, 0., 0., NULL, 0, 0};
extern unsigned long villageID;
extern double drand48();

Country *newCountry(void) {
	Country *newC;
	if (!freeC) {
		int i;
		freeC = (Country *)malloc(sizeof(Country) * AllocUnit);
		if (!freeC) { perror("malloc for new country"); exit(0); }
		memset(freeC, 0, sizeof(Country) * AllocUnit);
		for (i = 0; i < AllocUnit-1; i ++)
			freeC[i].post = &freeC[i + 1];
		freeC[AllocUnit-1].post = NULL;
	}
	newC = freeC;
	freeC = freeC->post;
	newC->pre = newC->post = NULL;
	newC->v = NULL;
	if (!newC->border) newC->border = newHash();
	newC->nVillages = 0;
	newC->id = ++ countryID;
	newC->age = newC->ageM = 0;
	newC->p = 0;
	newC->sv = newC->richness = 0.;
	newC->mil = 0;
	newC->logIndex = 0;
	setRandomMeme(newC->meme, sizeof(newC->meme));
	return newC;
}
releaseAllCountries(void) {
	if (theCountries) {
		Country *p, *last;
		for (p = theCountries; p; p = p->post) {
			 last = p;
			 releaseVillages(p->v);
			 releaseHashItems(p->border);
		}
		last->post = freeC;
		freeC = theCountries;
		theCountries = NULL;
	}
	countryID = villageID = 0;
	Step = 0;
	nCountries = 0;
}
void registarCountry(Country *c) {
	c->post = theCountries;
	if (theCountries) theCountries->pre = c;
	theCountries = c;
	nCountries ++;
	if (c->id == TracedCountryID)
		printf("%d:C%d was born.\n",Step,c->id);
}
void ruinCountry(Country *c) {
	if (c->pre) c->pre->post = c->post;
	else theCountries = c->post;
	if (c->post) c->post->pre = c->pre;
	releaseHashItems(c->border);
	c->post = freeC;
	freeC = c;
	nCountries --;
	if (c->id == TracedCountryID)
		printf("%d:C%d was ruined.\n",Step,c->id);
}
static void forAllCountries(void (*proc)(Country *)) {
	Country *c, *cc;
	for (c = theCountries; c; c = cc) {
		cc = c->post;
		(*proc)(c);
	}
}
static void forAllVillages(Country *c, void (*proc)(Village *)) {
	Village *v;
	for (v = c->v; v; v = v->post)
		(*proc)(v);
}
// Meme interpretation
static int MutationInterval(Country *c) {
	return c->meme[MemeCMutIntvl] / 5 + 1;
}
static double DeliveryRate(Country *c) {
	double v = c->meme[MemeCDelivery] * 1.2 / 255.;
	return (v < 1.)? v : 1.;
}
#define WeightDefence(c)	(c->meme[MemeCWd] / 255.)
#define WeightOffence(c)	(c->meme[MemeCWo] / 255.)
#define WeightPotential(c)	(c->meme[MemeCWp] / 255.)
#define BiasDefence(c)	(c->meme[MemeCBd] * 1.5 / 255.)
#define BiasOffence(c)	(c->meme[MemeCBo] * 2.5 / 255.)
#define BiasPotential(c)	(c->meme[MemeCBp] / 255.)
#define ThMilRatio(c)	(c->meme[MemeCMilRatio] * 0.6 / 255. + 0.2)
#define ThInvade(c)	(c->meme[MemeCInvade] / 255.)

// ACTION of country
static void reviseAge(Country *c) { c->age ++; }
static void regularMutation(Country *c) {
	if ((++ c->ageM) >= MutationInterval(c)) {
		mutatedCopy(c->meme, c->meme, sizeof(c->meme));
		c->ageM = 0;
	}
	forAllVillages(c, regularMutationVillage);
}
static void storeCLog(Country *c) {
	CountryLog *cl = &c->log[c->logIndex];
	cl->sv = c->sv;
	cl->p = c->p;
	cl->nv = c->nVillages;
	memcpy(cl->meme, c->meme, sizeof(c->meme));
	c->logIndex = (c->logIndex + 1) % CLogLength;
}
static void deliveryFoods(Country *c) {
	Village *v;
	double	fdSeed, fdNeed, pfs, pfn;
	fdSeed = fdNeed = 0.;
	for (v = c->v; v; v = v->post) {
		calcSurplus(v);
		if (v->surplus > 0.) fdSeed += v->surplus;
		else fdNeed -= v->surplus;
	}
	if (fdNeed < fdSeed) { pfs = fdNeed / fdSeed; pfn = 1.; }
	else if (fdNeed > fdSeed) { pfs = 1.; pfn = fdSeed / fdNeed; }
	else pfs = pfn = 0.;
	for (v = c->v; v; v = v->post) {
		v->fd = v->surplus * ((v->surplus > 0.)? pfs : pfn)
			* DeliveryRate(c);
		deliveryFoodsForVillage(v);
	}
}
static void calcCountryRichness(Country *c) {
	Village *v;
	c->p = 0; c->sv = 0.;
	for (v = c->v; v; v = v->post) {
		if (v->p < 0) {
			printf("calcCountryRichness:C%d,V%d. p=%d\n",v->p);
			exit(0);
		}
		c->p += v->p;
		c->sv += v->sv;
	}
	c->richness = (c->p > 0)? c->sv / c->p : 0.;
	storeCLog(c);
}
static unsigned long SmallestMil, LargestMil;
static void findMinMaxMil(void *item, void *udata) {
	Village *v = ((Land *)item)->v;
	if (v) {
		Country *c = v->k;
		if (SmallestMil > c->mil) SmallestMil = c->mil;
		if (LargestMil < c->mil) LargestMil = c->mil;
	}
}
static void setupPeopleJob(Country *c) {
	Village *v;
	double	ps, po;
	double	w1, w2, w3, wsum;
	if (c->p <= 0) return;
	c->work = 0;
	c->warDead = 0;
	c->nVindependent = c->nVmoveOut = c->nVmoveIn = 0;
//
	po = OfficerA + OfficerB / c->p;
	if (po > 1.) po = 1.;
//
	SmallestMil = ULONG_MAX; LargestMil = 0;
	forAllHashItems(findMinMaxMil, NULL, c->border);
	if (SmallestMil == ULONG_MAX) SmallestMil = 0;
	ps = c->p - (FoodCosumingPerPerson * c->p - c->sv)
		/ FoodProductionPerWorker;
	w1 = WeightDefence(c);
	w2 = WeightOffence(c);
	w3 = WeightPotential(c);
	wsum = w1 + w2 + w3;
	if (wsum > 0.) {
		ps = (w1 * BiasDefence(c) * LargestMil
		    + w2 * BiasOffence(c) * SmallestMil
		    + w3 * BiasPotential(c) * ps) / wsum;
		if (ps < 0.) ps = 0.;
		else if (ps > c->p) ps = c->p;
	} else ps = 0.;
//
	for (v = c->v; v; v = v->post) {
		v->po = v->p * po;
		v->ps = v->p * ps / c->p;
		if (v->ps + v->po >= v->p)
			v->ps = v->p - v->po;
		v->pp = v->p - v->po - v->ps;
	if (v->p < 0 || v->pp < 0 || v->ps < 0 || v->po < 0) {
		printf("setupPeopleJob:C%dV%d p=%d,(o:%d,s:%d,p:%d)\n",
			c->id,v->id,v->p,v->po,v->ps,v->pp);
		exit(0);
	}
	}
}
static void setupNewMil(Country *c) {
	Village *v;
	c->mil = 0;
	for (v = c->v; v; v = v->post) c->mil += v->ps;
}
static void decideVillageActionFor(Country *c) {
	forAllVillages(c, decideVillageAction);
}
struct targetRec { Country *offence; Village *target; };
static void resetVWorkFlagInLand(void *item, void *udata) {
	Village *v = ((Land *)item)->v;
	if (v) v->work = 0;
}
static void selectTarget(void *item, void *udata) {
	struct targetRec *tr = (struct targetRec *)udata;
	Village *v = ((Land *)item)->v;
	if (!v || !v->p || v->work) return;
	v->work = 1;
	if (v->k->mil / tr->offence->mil < ThMilRatio(tr->offence)
	 && v->richness >= tr->offence->richness
	 && (!tr->target || v->richness > tr->target->richness))
		tr->target = v;
}
static double triangleRandom() {
	double r = drand48();
	return (r < .5)? sqrt(r / 2.) : (1. - sqrt((1. - r) / 2.));
}
static void warDeadInCountry(Country *c, unsigned long m) {
	int	i, n, d;
	Village *v, **vs;
	d = triangleRandom() * m;
	for (v = c->v, n = 0; v; v = v->post) {
		int k = d * v->p / c->p;
		if (v->ps < k) k = v->ps;
		v->p -= k; v->ps -= k; n += k;
	}
	c->p -= d; c->mil -= d; c->warDead += d;
	if (n == d) return;
	vs = (Village **)alloca(sizeof(void *) * c->nVillages);
	for (i = 0, v = c->v; i < c->nVillages && v; i ++, v = v->post)
		vs[i] = v;
	if (i < c->nVillages) {
		printf("C%d %d %d\n",c->id,i,c->nVillages);
		exit(0);
	}
	randomShuffle((void *)vs, c->nVillages);
	for (i = 0; n < d; i = (i + 1) % c->nVillages) if (vs[i]->ps > 0) {
		vs[i]->ps --; vs[i]->p --; n ++;
	}
}
static void warDead(Country *co, Country *cd) {
	unsigned long m = (co->mil < cd->mil)? co->mil : cd->mil;
	warDeadInCountry(co, m);
	warDeadInCountry(cd, m);
}
static void attack(Country *co, Village	*v) {
	Country *cd = v->k;
	int	occupyP = 1;
	v->attacked = Attacked;
	nFights ++;
	if (co->id == TracedCountryID)
		printf("%d:C%d invaded V%d of C%d,",
			Step,co->id,v->id,cd->id);
	else if (cd->id == TracedCountryID)
		printf("%d:C%d was invaded to V%d by C%d,",
			Step,cd->id,v->id,co->id);
	if (cd->mil > 0) {
		double x = FightingBias * co->mil / cd->mil;
		double po = x * x / (1. + x * x);
		occupyP = (drand48() < po);
		warDead(co, cd);
	}
	if (co->id == TracedCountryID)
		printf(occupyP? " and won.\n" : " but failed.\n");
	else if (cd->id == TracedCountryID)
		printf(occupyP? " and lost.\n" : " but defended it.\n");
	if (!occupyP) return;
	nOccupation ++;
	moveCountry(v, cd, co);
	v->attacked = AttackedThenOccupied;
}
static void invadeIfItWishes(Country *c) {
	struct targetRec tr;
	if (c->p == 0 || c->mil == 0 ||
		drand48() < ThInvade(c)) return;
	tr.offence = c; tr.target = NULL;
	forAllHashItems(resetVWorkFlagInLand, NULL, c->border);
	forAllHashItems(selectTarget, (void *)(&tr), c->border);
	if (tr.target) attack(c, tr.target);
}
static forAllCountriesInRandomOrder(void (*proc)(Country *)) {
	int	i, n;
	Country	*c, **cs;
	for (n = 0, c = theCountries; c; c = c->post) n ++;
	cs = (Country **)allocTempMem(sizeof(void *) * n);
	for (i = 0, c = theCountries; c; c = c->post)
		cs[i ++] = c;
	randomShuffle((void *)cs, n);
	for (i = 0; i < n; i ++) if (cs[i]->v) (*proc)(cs[i]);
}
static unsigned long nVillages;
static Village **vList, **vListP;
static void countVillages(Country *c) {
	nVillages += c->nVillages;
}
static void gatherVillagesSub(Village *v) { *(vListP ++) = v; }
static void gatherVillages(Country *c) {
	forAllVillages(c, gatherVillagesSub);
}
static void forAllVillagesInRandomOrder(void (*proc)(Village *)) {
	int	i;
	nVillages = 0;
	forAllCountries(countVillages);
	vList = (Village **)allocTempMem(sizeof(void *) * nVillages);
	vListP = vList;
	forAllCountries(gatherVillages);
	randomShuffle((void *)vList, nVillages);
	for (i = 0; i < nVillages; i ++)
		(*proc)(vList[i]);
}
static void resetWorkFlag(Village *v) { v->work = 0; }
static unsigned char CountryWorkID;
static void traverseVillagesSub(void *item, void *udata) {
	static void traverseVillages(Village *v);
	Land *l = (Land *)item;
	Country *c = (Country *)udata;
	if (l->v && l->v->k == c && !l->v->work)
		traverseVillages(l->v);
}
static void traverseVillages(Village *v) {
	v->work = CountryWorkID;
	forAllHashItems(traverseVillagesSub, (void *)v->k, v->border);
}
static void checkCountrySep(Village *v) {
	if (!v->work) {
		CountryWorkID ++;
		traverseVillages(v);
	}
}
static void checkSepAndRuin(Country *c) {
	extern void checkSepCountryBorder(void *, void *);
	Village *v, *vv;
	Country **cs, *cc;
	int	i, myID, maxNv, *nv;
	struct { int n; Country **cs; } csn;
	if (c->id == TracedCountryID) {
		if (c->warDead)
			printf("%d:C%d %d were died in war.\n",
				Step,c->id,c->warDead);
		if (c->nVindependent || c->nVmoveOut || c->nVmoveIn) {
			printf("%d:C%d %d villeges",Step,c->id,c->nVillages);
			if (c->nVindependent > 0)
				printf(", %d got independent",c->nVindependent);
			if (c->nVmoveOut > 0)
				printf(", %d moved out",c->nVmoveOut);
			if (c->nVmoveIn > 0)
				printf(", %d moved in",c->nVmoveIn);
			printf(".\n");
		}
	}
	if (!c->v) { ruinCountry(c); return; }
	if (!c->work) return;
	forAllVillages(c, resetWorkFlag);
	CountryWorkID = 0;
	forAllVillages(c, checkCountrySep);
	if (CountryWorkID <= 1) return;
	cs = (Country **)alloca(sizeof(void *) * CountryWorkID * 2);
	nv = (int *)(cs + CountryWorkID);
	memset(nv,0,sizeof(int)*CountryWorkID);
	for (v = c->v; v; v = v->post) nv[v->work - 1] ++;
	myID = 0; maxNv = nv[0];
	for (i = 1; i < CountryWorkID; i ++)
		if (maxNv < nv[i]) { maxNv = nv[i]; myID = i; }
	cs[myID] = c;
	for (i = 0; i < CountryWorkID; i ++) if (i != myID) {
		cs[i] = newCountry();
		memcpy(cs[i]->meme,c->meme,sizeof(c->meme));
		cs[i]->ageM = c->ageM;
		registarCountry(cs[i]);
		setSimilarRGB(&cs[i]->rgb, &c->rgb);
	}
	for (v = c->v; v; v = vv) {
		vv = v->post;
		if (v->work != myID + 1) {
			if (v->pre) v->pre->post = vv;
			else c->v = vv;
			if (vv) vv->pre = v->pre;
			cc = cs[v->work - 1];
			v->pre = NULL;
			v->post = cc->v;
			if (v->post) v->post->pre = v;
			cc->v = v;
			v->k = cc;
			c->nVillages --;
			cc->nVillages ++;
		}
	}
	csn.n = CountryWorkID;
	csn.cs = cs;
	cs[myID] = cs[0]; cs[0] = c;
	forAllHashCells(checkSepCountryBorder, (void *)(&csn), c->border);
	if (c->id == TracedCountryID) {
		printf("%d:C%d separated into",Step,c->id);
		for (i = 0; i < CountryWorkID; i ++)
			printf("%cC%d(%d)",(i?',':' '),
				cs[i]->id,cs[i]->nVillages);
		printf("\n");
	}
	for (i = 0; i < CountryWorkID; i ++)
		if (cs[i]->nVillages <= 0) break;
	if (i >= CountryWorkID) return;
	for (i = 0; i < CountryWorkID; i ++) {
		char	c = '{';
		printf("  C%d = ",cs[i]->id);
		if (!cs[i]->v) printf("{");
		else for (v = cs[i]->v; v; v = v->post) {
			printf("%c%d",c,v->id); c = ',';
		}
		printf("}, nv=%d\n",cs[i]->nVillages);
	}
	exit(0);
}
static unsigned long nTerritorialLands;
static unsigned long logMaxVillages, logSumVillages;
static unsigned long sumPop_p, sumPop_s, sumPop_o;
static void sumUpPopulation(Village *v) {
	sumPop_p += v->pp;
	sumPop_s += v->ps;
	sumPop_o += v->po;
	nPopulation += v->p;
	nTerritorialLands += v->nLands;
}
static void setCountryLogDataSub(Country *c) {
	if (c->nVillages > 100000)
		printf("Country #%d includes %d villages.\n",
			c->nVillages);
	if (logMaxVillages < c->nVillages)
		logMaxVillages = c->nVillages;
	logSumVillages += c->nVillages;
	forAllVillages(c, sumUpPopulation);
	nWarDead += c->warDead;
}
static void setLogData(LogData *ld, double *x) {
	int	i;
	if (ld->size <= ld->index) {
		int ns = (ld->size / AllocUnit + 1) * AllocUnit;
		ld->data = (double *)realloc(ld->data,
			sizeof(double) * ns * ld->n);
		if (!ld->data) { perror("realloc in setLogData"); exit(0); }
		ld->size = ns;
	}
	for (i = 0; i < ld->n; i ++) {
		ld->data[ld->index ++] = x[i];
		if (ld->min > x[i]) ld->min = x[i];
		if (ld->max < x[i]) ld->max = x[i];
	}
}
static void setCountryLogData(void) {
	double data[3];
	nPopulation = nTerritorialLands = 0;
	logMaxVillages = logSumVillages = 0;
	sumPop_p = sumPop_s = sumPop_o = 0;
	forAllCountries(setCountryLogDataSub);
	data[0] = (double)sumPop_s;
	data[1] = (double)(sumPop_s + sumPop_o);
	data[2] = (double)(sumPop_s + sumPop_o + sumPop_p);
	setLogData(&logPopulation, data);
	data[0] = (double)nTerritorialLands;
	setLogData(&logTerritory, data);
	data[0] = (double)nCountries;
	setLogData(&logNCountries, data);
	data[0] = (double)logMaxVillages;
	setLogData(&logMaxNVillages, data);
	data[0] = (double)logSumVillages / nCountries;
	setLogData(&logAveNVillages, data);
	data[0] = (double)nFights / nCountries;
	data[1] = (double)nOccupation / nCountries;
	setLogData(&logNFights, data);
	data[0] = (double)nWarDead / nPopulation;
	setLogData(&logWarDead, data);
	showTrend();
}
void resetCountryLog(void) {
	static LogData *logs[] = {
		&logPopulation, &logTerritory, &logNCountries,
		&logMaxNVillages, &logAveNVillages,
		&logNFights, &logWarDead,
		NULL
	};
	LogData	**p;
	for (p = logs; *p; p ++) {
		(*p)->min = (*p)->max = 0.;
		(*p)->index = (*p)->size = 0ul;
	}
	setCountryLogData();
}
void doOneStep(void) {
	Step ++;
	nFights = nOccupation = nPopulation = nWarDead = 0;
	forAllCountries(reviseAge);
	if (FunctionSwitch & FnRegularMutate)
		forAllCountries(regularMutation);
	if (FunctionSwitch & FnDelivery)
		forAllCountries(deliveryFoods);
	forAllCountries(calcCountryRichness);
	forAllCountries(setupPeopleJob);
	forAllCountries(setupNewMil);
	forAllCountries(decideVillageActionFor);
	if (FunctionSwitch & FnWar)
		forAllCountriesInRandomOrder(invadeIfItWishes);
	forAllVillagesInRandomOrder(doVillageAction);
	forAllCountries(checkSepAndRuin);
	setCountryLogData();
}
void showCountry(Country *c) {
	Village *v;
	printf("Country #%d\n", c->id);
	printf("  Villages: %d, Population: %d, FoodStock: %g, Richness: %g\n",
		c->nVillages, c->p, c->sv, c->richness);
}
void showCountryLog(Country *c) {
	int	i, j, n;
	if (c->age < CLogLength) {
		j = 0; n = c->age;
	} else {
		j = c->logIndex; n = CLogLength;
	}
	printf("Country #%d (%d steps old)\n",c->id,c->age);
	for (i = 0; i < n; i ++, j = (j + 1) % CLogLength) {
		CountryLog *cl = &c->log[j];
		printf("%d:nv=%d,p=%d,sv=%g,",Step-n+i,
			cl->nv,cl->p,cl->sv);
		printf("D=%g,Tm=%g,Ti=%g\n",
			cl->meme[MemeCDelivery] * 1.2 / 255.,
			ThMilRatio(cl),ThInvade(cl));
	}
}
static void showCountryShort(Country *c) {
	printf("%d %d %d %g %g %d\n", c->id,
		c->nVillages, c->p, c->sv, c->richness,
		c->warDead);
}
showAllCountries(void) {
	forAllCountries(showCountryShort);
}
