#include <Xm/MainW.h>
#include <Xm/CascadeB.h>
#include <Xm/DrawingA.h>
#include <Xm/Label.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/Separator.h>
#include <Xm/Text.h>
#include <Xm/ToggleB.h>
#include "war.h"

#define	PrevMapUnit	4

static XtAppContext	app_con;
Display	*display;
GC	gc;
Widget MainWindow;	
static Window window, msgwin;
static Pixel fg, bg;
static Pixmap pmap;
static Dimension width, height, msgheight;
static Widget app_sh, startBtn, stopBtn, stepBtn;
static int Running = FALSE;
static int SleepTime = 10;
static unsigned char *prevMap;
enum {
	ViewMapBoth, ViewMapVillage, ViewMapVillagePh,
	ViewMapCountry, ViewMapCountryPh,
	ViewMapBorder
};
static int viewMode = ViewMapBoth;
static unsigned long ClickedCountryID = 0, ClickedVillageID = 0;
enum {
	DrawVillageBorder = 1, DrawCountryBorder = 2
};
static int drawWhatBorder = DrawCountryBorder;

extern int areaWidth, areaHeight;
extern Land *theLand;
extern void doOneStep();
extern void setupMenuItemTrend(Widget);
extern void setupMenuItemParam(Widget);

static void TimeoutProc(cl, id) XtPointer cl; XtIntervalId *id; {
	extern Country *theCountries;
	if (theCountries) {
		doOneStep();
		showMap();
	} else	Running = False;
	if (Running) XtAppAddTimeOut(app_con,SleepTime,TimeoutProc,NULL);
}
static Boolean WorkProc(XtPointer cl) {
	void (*proc)(void) = (void (*)(void))cl;
	(*proc)();
	return True;
}
void reserveProc(void (*proc)(void)) {
	XtAppAddWorkProc(app_con,WorkProc,(XtPointer)proc);
}
static void clearBackMap(void) {
	memset(prevMap, 0, areaWidth * areaHeight * PrevMapUnit); 
	XSetForeground(display,gc,BlackPixel(display,0));
	XFillRectangle(display,pmap,gc,0,0,width,height);
}
static void QuitCB(w,cl,cd) Widget w; XtPointer cl,cd; { exit(0); }
static void viewModeChanged(void) {
		clearBackMap();
		showMap();
}	
static void ViewChangeCB(w,cl,cd) Widget w; XtPointer cl,cd; {
	int newMode = (int)cl;
	Boolean	value;
	XtVaGetValues(w,XmNset,&value,NULL);
	if (value && viewMode != newMode) {
		viewMode = newMode;
		reserveProc(viewModeChanged);
	}
}
static void ResetCB(w,cl,cd) Widget w; XtPointer cl,cd; {
	reserveProc(resetWorld);
}
static void RestartCB(w,cl,cd) Widget w; XtPointer cl,cd; {
	reserveProc(restartWorld);
}
static void StartCB(w,cl,cd) Widget w; XtPointer cl,cd; {
	Running = TRUE;
	XtVaSetValues(startBtn,XmNsensitive,FALSE,NULL);
	XtVaSetValues(stepBtn,XmNsensitive,FALSE,NULL);
	XtVaSetValues(stopBtn,XmNsensitive,TRUE,NULL);
	XtAppAddTimeOut(app_con,SleepTime,TimeoutProc,NULL);
}
static void StepCB(w,cl,cd) Widget w; XtPointer cl,cd; {
	XtAppAddTimeOut(app_con,SleepTime,TimeoutProc,NULL);
}
static void StopCB(w,cl,cd) Widget w; XtPointer cl,cd; {
	Running = FALSE;
	XtVaSetValues(startBtn,XmNsensitive,TRUE,NULL);
	XtVaSetValues(stepBtn,XmNsensitive,TRUE,NULL);
	XtVaSetValues(stopBtn,XmNsensitive,FALSE,NULL);
}
static Widget digitsText;
static void OkCB(w,cl,cd) Widget w; XtPointer cl,cd; {
	unsigned long *id = (unsigned long *)cl;
	char *digits;
	XtVaGetValues(digitsText,XmNvalue,&digits,NULL);
	*id = atoi(digits);
	printf("tracing %d\n",*id);
}
static void setTrace(char *name, unsigned long *id, unsigned long dflt) {
	static Widget dialog = NULL;
	char	buf[48];
	XmString	msg;
	if (!dialog) {
		dialog = XmCreateQuestionDialog(app_sh,"traceDialog",NULL,0);
		XtUnmanageChild(XmMessageBoxGetChild(dialog,XmDIALOG_HELP_BUTTON));
		XtManageChild(digitsText=XmCreateText(dialog,"idText",NULL,0));
		XtAddCallback(dialog,XmNokCallback,OkCB,(XtPointer)id);
	}
	sprintf(buf,"Input ID of traced %s.",name);
	msg = XmStringCreateSimple(buf);
	XtVaSetValues(dialog,XmNmessageString,msg,NULL);
	sprintf(buf,"%d",dflt);
	XtVaSetValues(digitsText,XmNvalue,buf,NULL);
	XtManageChild(dialog);
	XmStringFree(msg);
}
static void TraceCountryCB(w,cl,cd) Widget w; XtPointer cl,cd; {
	setTrace("country", &TracedCountryID, ClickedCountryID);
}
static void TraceVillageCB(w,cl,cd) Widget w; XtPointer cl,cd; {
	setTrace("village", &TracedVillageID, ClickedVillageID);
}
enum { DEBUGshowAllCountries };
static void DebugCB(w,cl,cd) Widget w; XtPointer cl, cd; {
	switch ((long)cl) {
		case DEBUGshowAllCountries: showAllCountries(); break;
	}
}
static void BorderTypeCB(w,cl,cd) Widget w; XtPointer cl, cd; {
	Boolean value;
	XtVaGetValues(w,XmNset,&value,NULL);
	if (value) drawWhatBorder |= (int)cl;
	else drawWhatBorder &= ~ (int)cl;
}
static void drawBorder1(void *item, void *udata) {
	int	w = width / areaWidth, h = height / areaHeight;
	Land *l = (Land *)item;
	XFillRectangle(display,window,gc,l->x*w,l->y*h,w,h);
}
static void drawBorder(void *border, Pixel pxl) {
	XSetForeground(display,gc,pxl);
	forAllHashItems(drawBorder1,NULL,border);
}
static void showLVC(Land *l) {
	extern void showLand(Land *);
	extern void showVillage(Village *);
	extern void showCountry(Country *);
	showLand(l);
	XCopyArea(display,pmap,window,gc,0,0,width,height,0,0);
	if (l->v) {
		showVillage(l->v);
		showCountry(l->v->k);
		if (drawWhatBorder & DrawVillageBorder)
			drawBorder(l->v->border,0x6666ff);
		if (drawWhatBorder & DrawCountryBorder)
			drawBorder(l->v->k->border,0xffffff);
		ClickedVillageID = l->v->id;
		ClickedCountryID = l->v->k->id;
	}
}
static void showCLog(Land *l) {
	extern void showCountryLog(Country *);
	if (!l->v) return;
	showCountryLog(l->v->k);
	ClickedCountryID = l->v->k->id;
}
static void ClickCB(w,cl,cd) Widget w; XtPointer cl;
	XmDrawingAreaCallbackStruct *cd; {
	XButtonEvent *ev = (XButtonEvent *)cd->event;
	if (ev->type == ButtonPress) {
		int x = ev->x / (width / areaWidth);
		int y = ev->y / (height / areaHeight);
		Land *l = &theLand[y * areaWidth + x];
		switch (ev->button) {
			case Button1: showLVC(l); break;
			case Button3: showCLog(l);
		}
	}
}
static void RedrawCB(w,cl,cd) Widget w; XtPointer cl;
	XmDrawingAreaCallbackStruct *cd; {
	XExposeEvent *ev = (XExposeEvent *)cd->event;
	XCopyArea(display,pmap,window,gc,
		ev->x,ev->y,ev->width,ev->height,ev->x,ev->y);
}
static void MsgRedrawCB(w,cl,cd) Widget w; XtPointer cl, cd; {
	char msg[32];
	extern unsigned long Step, nCountries;
	sprintf(msg,"%6d %4d countries",Step,nCountries);
	XSetForeground(display,gc,fg);
	XDrawImageString(display,msgwin,gc,10,msgheight - 4,
		msg,strlen(msg));
}
initUserIF(int argc, char *argv[]) {
	Widget menu_bar, menu, button, draw, msg;
	int depth;
	Arg arg[1];
	extern String fallback_resources[];
	app_sh = XtAppInitialize(&app_con,"War",NULL,0,
		&argc,argv,fallback_resources,NULL,0);
	XtManageChild(MainWindow=XmCreateMainWindow(app_sh,"MainWindowdow",NULL,0));
	XtManageChild(menu_bar=XmCreateMenuBar(MainWindow,"menu_bar",NULL,0));

	menu = XmCreatePulldownMenu(app_sh,"FileMenu",NULL,0);
	XtManageChild(button=XmCreatePushButton(menu,"Restart",NULL,0));
	XtAddCallback(button,XmNactivateCallback,RestartCB,NULL);
	XtManageChild(button=XmCreatePushButton(menu,"Reset",NULL,0));
	XtAddCallback(button,XmNactivateCallback,ResetCB,NULL);
	XtManageChild(button=XmCreatePushButton(menu,"Quit",NULL,0));
	XtAddCallback(button,XmNactivateCallback,QuitCB,NULL);
	XtSetArg(arg[0],XmNsubMenuId,menu);
	XtManageChild(button=XmCreateCascadeButton(menu_bar,"File",arg,1));
	XtSetArg(arg[0],XmNradioBehavior,True);

	menu = XmCreatePulldownMenu(app_sh,"ViewMenu",arg,1);
	XtSetArg(arg[0],XmNset,True);
	XtManageChild(button=XmCreateToggleButton(menu,"Both",arg,1));
	XtAddCallback(button,XmNvalueChangedCallback,
		ViewChangeCB,(XtPointer)ViewMapBoth);
	XtManageChild(button=XmCreateToggleButton(menu,"Village patch",NULL,0));
	XtAddCallback(button,XmNvalueChangedCallback,
		ViewChangeCB,(XtPointer)ViewMapVillage);
	XtManageChild(button=XmCreateToggleButton(menu,"Village philogeny",NULL,0));
	XtAddCallback(button,XmNvalueChangedCallback,
		ViewChangeCB,(XtPointer)ViewMapVillagePh);
	XtManageChild(button=XmCreateToggleButton(menu,"Country patch",NULL,0));
	XtAddCallback(button,XmNvalueChangedCallback,
		ViewChangeCB,(XtPointer)ViewMapCountry);
	XtManageChild(button=XmCreateToggleButton(menu,"Country philogeny",NULL,0));
	XtAddCallback(button,XmNvalueChangedCallback,
		ViewChangeCB,(XtPointer)ViewMapCountryPh);
	XtManageChild(button=XmCreateToggleButton(menu,"Border",NULL,0));
	XtAddCallback(button,XmNvalueChangedCallback,
		ViewChangeCB,(XtPointer)ViewMapBorder);
	XtSetArg(arg[0],XmNsubMenuId,menu);
	XtManageChild(XmCreateCascadeButton(menu_bar,"Color",arg,1));

	XtManageChild(startBtn=XmCreateCascadeButton(menu_bar,"Start",NULL,0));
	XtAddCallback(startBtn,XmNactivateCallback,StartCB,NULL);
	XtManageChild(stepBtn=XmCreateCascadeButton(menu_bar,"Step",NULL,0));
	XtAddCallback(stepBtn,XmNactivateCallback,StepCB,NULL);
	XtSetArg(arg[0],XmNsensitive,FALSE);
	XtManageChild(stopBtn=XmCreateCascadeButton(menu_bar,"Stop",arg,1));
	XtAddCallback(stopBtn,XmNactivateCallback,StopCB,NULL);
	
	menu = XmCreatePulldownMenu(app_sh,"WindowMenu",NULL,0);
	setupMenuItemTrend(menu);
	setupMenuItemParam(menu);
	XtSetArg(arg[0],XmNsubMenuId,menu);
	XtManageChild(XmCreateCascadeButton(menu_bar,"Window",arg,1));

	menu = XmCreatePulldownMenu(app_sh,"TraceMenu",NULL,0);
	XtManageChild(button=XmCreatePushButton(menu,"Country ...",NULL,0));
	XtAddCallback(button,XmNactivateCallback,TraceCountryCB,NULL);
	XtManageChild(button=XmCreatePushButton(menu,"Village ...",NULL,0));
	XtAddCallback(button,XmNactivateCallback,TraceVillageCB,NULL);
	XtSetArg(arg[0],XmNsubMenuId,menu);
	XtManageChild(XmCreateCascadeButton(menu_bar,"Trace",arg,1));

	menu = XmCreatePulldownMenu(app_sh,"DebugMenu",NULL,0);
	XtManageChild(button=XmCreatePushButton(menu,"Show all countries",NULL,0));
	XtAddCallback(button,XmNactivateCallback,DebugCB,
		(XtPointer)DEBUGshowAllCountries);
	XtSetArg(arg[0],XmNset,((drawWhatBorder & DrawVillageBorder) != 0));
	XtManageChild(button=XmCreateToggleButton(menu,"Village border",arg,1));
	XtAddCallback(button,XmNvalueChangedCallback,BorderTypeCB,
		(XtPointer)DrawVillageBorder);
	XtSetArg(arg[0],XmNset,((drawWhatBorder & DrawCountryBorder) != 0));
	XtManageChild(button=XmCreateToggleButton(menu,"Country border",arg,1));
	XtAddCallback(button,XmNvalueChangedCallback,BorderTypeCB,
		(XtPointer)DrawCountryBorder);
	XtSetArg(arg[0],XmNsubMenuId,menu);
	XtManageChild(XmCreateCascadeButton(menu_bar,"Debug",arg,1));

	XtManageChild(draw=XmCreateDrawingArea(MainWindow,"field",NULL,0));
	XtManageChild(msg=XmCreateDrawingArea(MainWindow,"message",NULL,0));
	XtVaSetValues(MainWindow,XmNmenuBar,menu_bar,XmNmessageWindow,msg,
		XmNshowSeparator,True,NULL);
	XtRealizeWidget(app_sh);
	window = XtWindow(draw);
	msgwin = XtWindow(msg);
	gc = DefaultGC(display = XtDisplay(app_sh), 0);
	XtVaGetValues(draw,XmNwidth,&width,XmNheight,&height,
		XmNdepth,&depth,XmNforeground,&fg,XmNbackground,&bg,NULL);
	pmap = XCreatePixmap(display,window,width,height,depth);
	XSetBackground(display,gc,bg);
	XtVaGetValues(msg,XmNheight,&msgheight,NULL);
	//printf("%d x %d\n",areaWidth,areaHeight);
	prevMap = (unsigned char *)malloc(areaWidth * areaHeight * PrevMapUnit);
	if (!prevMap) { perror("malloc for prevMap"); exit(0); }
	clearBackMap();
	XtAddCallback(draw,XmNinputCallback,ClickCB,NULL);
	XtAddCallback(draw,XmNexposeCallback,RedrawCB,NULL);
	XtAddCallback(msg,XmNexposeCallback,MsgRedrawCB,NULL);
}
loopUserIF(void) {
	XtAppMainLoop(app_con);
}

#define NColors 64
static Pixel colors[] = {
 0x000000,
 0xffffff, 0xffff00, 0xff00ff, 0x00ffff,
 0xff8800, 0x0088ff, 0x88ff00, 0xff0088,
 0x8800ff, 0x00ff88, 0xff4444, 0x44ff44,
 0x4444ff, 0xff4488, 0x4488ff, 0x88ff44,
 0xff8844, 0x8844ff, 0x44ff88, 0xcccc44,
 0xcc44cc, 0x44cccc, 0xcc44ff, 0x44ffcc,
 0xffcc44, 0x44ccff, 0xccff44, 0xff44cc,
 0x888800, 0x008888, 0x880088, 0x884400,
 0x440088, 0x008844, 0x880044, 0x004488,
 0x448800, 0x444444, 0xcc0044, 0x0044cc,
 0x44cc00, 0x4400cc, 0x00cc44, 0xcc4400,
 0x880000, 0x008800, 0x000088, 0x444400,
 0x440044, 0x004444, 0xcc4444, 0x44cc44,
 0x4444cc, 0xcccccc, 0xff00cc, 0x00ccff,
 0xccff00, 0xffcc00, 0xcc00ff, 0xffcc00,
 0xff0000, 0x00ff00, 0x0000ff, 0x888888
}; 

static void drawOneGrid(int i, int j, int w, int h, int id1, int id2) {
	int	idx = (j * areaWidth + i) * PrevMapUnit, cid1, cid2;
	cid1 = (id1 < 0)? 0 : (id1 % NColors + 1);
	cid2 = (id2 < 0)? cid1 : (id2 % NColors + 1);
	if (prevMap[idx] != cid1 || prevMap[idx+1] != cid2) {
		XSetForeground(display,gc,colors[cid1]);
		XFillRectangle(display,pmap,gc,j*w,i*h,w,h);
		prevMap[idx] = cid1;
		if (cid1 != cid2) {
			int ww = (w > 2)? w - 2 : 1, hh = (h > 2)? h - 2 : 1;
			XSetForeground(display,gc,colors[cid2]);
			XFillRectangle(display,pmap,gc,j*w+1,i*h+1,ww,hh);
		}
		prevMap[idx + 1] = cid2;
	}
}
static void drawOneGridPh(int i, int j, int w, int h, MyRGB *c) {
	Pixel pxl = c? (c->r << 16) + (c->g << 8) + c->b : 0;
	int idx = (j * areaWidth + i) * PrevMapUnit;
	if (*(Pixel *)(&prevMap[idx]) != pxl) {
		XSetForeground(display,gc,pxl);
		XFillRectangle(display,pmap,gc,j*w,i*h,w,h);
		*(Pixel *)(&prevMap[idx]) = pxl;
	}
}
static void drawOneGridBorder(int i, int j, int w, int h,
	int dx, int dy, Village *v) {
	static Pixel pixels[] = { 0x000000, 0x00ccee, 0xffffff };
	int idx = (j * areaWidth + i) * PrevMapUnit + dy;
	int k = ((i + dy) % areaHeight) * areaWidth + (j + dx) % areaWidth;
	Village *vv = theLand[k].v;
	int cid = (v == vv)? 0 : ((!v || !vv || v->k != vv->k)? 2 : 1);
	if (prevMap[idx] != cid) {
		XSetForeground(display,gc,pixels[cid]);
		XDrawLine(display,pmap,gc,
			(j+dx)*w+dy,(i+dy)*h+dx,(j+1)*w,(i+1)*h);
		prevMap[idx] = cid;
	}
}
static void drawMapGrid(Village *v, int i, int j, int w, int h) {
	switch (viewMode) {
		case ViewMapBorder:
		drawOneGridBorder(i,j,w,h,1,0,v);
		drawOneGridBorder(i,j,w,h,0,1,v);
		break;
		case ViewMapVillagePh:
		drawOneGridPh(i,j,w,h,(v? &v->rgb : NULL));
		break;
		case ViewMapCountryPh:
		drawOneGridPh(i,j,w,h,(v? &v->k->rgb : NULL));
		break;
		default:
		if (!v) drawOneGrid(i,j,w,h,-1,-1);
		else switch (viewMode) {
			case ViewMapBoth: drawOneGrid(i,j,w,h,v->k->id,v->id);
			break;
			case ViewMapVillage: drawOneGrid(i,j,w,h,v->id,-1);
			break;
			case ViewMapCountry: drawOneGrid(i,j,w,h,v->k->id,-1);
		}
	}
}
showMap(void) {
	int i, j, w = width / areaWidth, h = height / areaHeight;
	for (i = 0; i < areaHeight; i ++)
	for (j = 0; j < areaWidth; j ++) {
		Land *p = &theLand[i * areaWidth + j];
		drawMapGrid(p->v, i, j, w, h);
	}
	XCopyArea(display,pmap,window,gc,0,0,width,height,0,0);
	MsgRedrawCB(NULL,NULL,NULL);
}
