#include <stdio.h>
#include "hashItem.h"
#define AllocUnit 32

static HashItem *freeHI = NULL;

static unsigned int hash(unsigned long key) {
	return (key * 104729ul) % HashSize;
}
void *newHash(void) {
	HashTable h = (HashTable)malloc(sizeof(void *) * HashSize);
	if (!h) { perror("malloc error for hash table."); exit(0); }
	memset(h, 0, sizeof(void *) * HashSize);
	return (void *)h;
}
void releaseHashItems(HashTable ht) {
	int i;
	for (i = 0; i < HashSize; i ++) {
		HashItem *h = ht[i];
		if (h) {
			while (h->post) h = h->post;
			h->post = freeHI;
			freeHI = ht[i];
			ht[i] = NULL;
		}
	}
}
void *getHash(long key, HashTable ht) {
	HashItem *h;
	for (h = ht[hash(key)]; h; h = h->post)
		if (h->key == key) return h->item;
	return NULL;
}
int putHash(long key, void *item, HashTable ht) {
	HashItem *h;
	int	k = hash(key);
	for (h = ht[k]; h; h = h->post)
		if (h->key == key) { h->item = item; return 0; }
	if (!freeHI) {
		int	i;
		freeHI = (HashItem *)malloc(sizeof(HashItem) * AllocUnit);
		if (!freeHI) { perror("malloc error for hash item."); exit(0); }
		for (i = 0; i < AllocUnit - 1; i ++)
			freeHI[i].post = &freeHI[i + 1];
		freeHI[AllocUnit - 1].post = NULL;
	}
	h = freeHI;
	freeHI = h->post;
	h->pre = NULL;
	h->post = ht[k];
	if (h->post) h->post->pre = h;
	ht[k] = h;
	h->key = key;
	h->item = item;
	return 1;
}
void removeHash(long key, HashTable ht) {
	HashItem **h;
	for (h = &ht[hash(key)]; *h; h = &((*h)->post))
		if ((*h)->key == key) break;
	if (*h) {
		HashItem *p = *h;
		if (p->post) p->post->pre = p->pre;
		*h = p->post;
		p->post = freeHI;
		freeHI = p;
	}
}
int countHashItems(HashTable ht) {
	int i, n;
	HashItem *h;
	for (i = n = 0; i < HashSize; i ++)
	for (h = ht[i]; h; h = h->post) n ++;
	return n;
}
void *randomItemInHash(int (*proc)(void *), HashTable ht) {
	void	**temp;
	int	i, n, k;
	HashItem	*h;
	n = countHashItems(ht);
	if (n <= 0) return NULL;
	temp = (void **)alloca(sizeof(void *) * n);
	for (i = n = 0; i < HashSize; i ++)
	for (h = ht[i]; h; h = h->post)
		if ((*proc)(h->item)) temp[n ++] = h->item;
	return (n > 0)? temp[lrand48() % n] : NULL;
}
void forAllHashItems(void (*proc)(void *, void *),
	void *udata, HashTable ht) {
	int	i;
	HashItem *h;
	for (i = 0; i < HashSize; i ++)
	for (h = ht[i]; h; h = h->post)
		(*proc)(h->item, udata);
}
void forAllHashCells(void (*proc)(void *, void *),
	void *udata, HashTable ht) {
	int	i;
	HashItem *h, *hh;
	for (i = 0; i < HashSize; i ++)
	for (h = ht[i]; h; h = hh) {
		hh = h->post;
		(*proc)((void *)h, udata);
	}
}
void removeHashCell(HashItem *h, HashTable ht) {
	if (h->pre) h->pre->post = h->post;
	else ht[hash(h->key)] = h->post;
	if (h->post) h->post->pre = h->pre;
}
