#include <stdio.h>
#include <math.h>
#include <Xm/DrawingA.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/Scale.h>
#include <Xm/Text.h>
#include <Xm/ToggleB.h>
#include "war.h"

extern Display	*display;
extern GC	gc;
extern Widget	MainWindow;
static Widget	Mshell = 0, trendBtn;
static Window	MWin;
static Pixmap	MPmap;
static Pixel	Mfg, Mbg;
static XFontStruct	*Gfont;
static unsigned int	Mwidth, Mheight;
#define	MXtitle	"step"
#define	MYoffset	20
static	int	MXoffset, MYspan, MXtitleX, MXtitleY,
		MXbase, MYbase;
static int	AverageWidth = 1;
extern void reserveProc(void (*)(void));
static LogData *currentViewLog;

static void ChangeViewCB(Widget w, XtPointer cl, XtPointer cd) {
	void showTrend(void);
	currentViewLog = (LogData *)cl;
	showTrend();
}
static void CancelCB(Widget w, XtPointer cl, XtPointer cd) {
	XtUnmanageChild(w);
}
static void SaveSelOKCB(Widget w, XtPointer cl, XtPointer cd) {
	LogData *ld = currentViewLog;
	FILE	*out;
	int	i, j;
	char *path;
	extern	unsigned long	Step;
	path = XmTextGetString(XmFileSelectionBoxGetChild(w,XmDIALOG_TEXT));
	out = fopen(path, "w");
	for (i = 0; i <= Step; i ++) {
		fprintf(out, "%d", i);
		for (j = 0; j < ld->n; j ++)
			fprintf(out, " %g", ld->data[i*ld->n + j]);
		putc('\n', out);
	}
	fclose(out);
	XtUnmanageChild(w);
}
static void SaveLogCB(Widget w, XtPointer cl, XtPointer cd) {
	static Widget fsb = 0;
	if (!fsb) {
		Widget helpBtn;
		fsb = XmCreateFileSelectionDialog(Mshell,"file_sel",NULL,0);
		helpBtn = XmFileSelectionBoxGetChild(fsb,XmDIALOG_HELP_BUTTON);
		XtUnmanageChild(helpBtn);
		XtAddCallback(fsb,XmNokCallback,SaveSelOKCB,NULL);
		XtAddCallback(fsb,XmNcancelCallback,CancelCB,NULL);
	}
	XtManageChild(fsb);
}
static void ScaleCB(Widget w, XtPointer cl, XtPointer cd) {
	static void drawLogData(LogData *);
	XmScaleCallbackStruct *cs = (XmScaleCallbackStruct *)cd;
	AverageWidth = cs->value;
	drawLogData(currentViewLog);
}
static void redisplayMonitorCB(Widget w, XtPointer cl, XtPointer cd) {
	XExposeEvent *e = (XExposeEvent *)
		((XmDrawingAreaCallbackStruct *)cd)->event;
	XCopyArea(display,MPmap,MWin,gc,
		e->x,e->y,e->width,e->height,e->x,e->y);
}
static void get_root_geometry(Widget w, int *x, int *y,
	unsigned int *width, unsigned int *height) {
	Window	window, parent = XtWindow(w), root = 0, *children;
	unsigned int	mw, mh, border, depth, n;
	do {	window = parent;
		XQueryTree(display,window,&root,&parent,&children,&n);
		XFree(children);
	} while (root != parent);
	XGetGeometry(display,window,&root,x,y,width,height,&border,&depth);
}
static void setupLogTypes(Widget menu) {
	extern LogData logPopulation, logTerritory, logNCountries,
		logMaxNVillages, logAveNVillages,
		logNFights, logWarDead;
	static struct { char *name; LogData *log; } ldata[] = {
		{"Population",&logPopulation},
		{"Territory",&logTerritory},
		{"Countries",&logNCountries},
		{"AvrgNVil",&logAveNVillages},
		{"MaxNVil",&logMaxNVillages},
		{"War",&logNFights},
		{"WarDead",&logWarDead},
		{NULL}
	}, *p;
	for (p = ldata; p->name; p ++) {
		Widget button = XmCreatePushButton(menu,p->name,NULL,0);
		XtManageChild(button);
		XtAddCallback(button,XmNactivateCallback,ChangeViewCB,
			(XtPointer)p->log);
	}
	currentViewLog = ldata[0].log;
}
static void createTrendShell(void) {
	Widget	main, menu, button, scale, work;
	Window	root;
	int	x, y, n;
	unsigned int	width,height,border,depth;
	XmFontList	fontList;
	XmFontContext	context;
	XmStringCharSet	charset;
	Arg	args[4];
	get_root_geometry(MainWindow,&x,&y,&width,&height);
	XtSetArg(args[0],XmNx,x+width);
	XtSetArg(args[1],XmNy,y);
	Mshell = XtCreatePopupShell("Monitor",
		transientShellWidgetClass,MainWindow,args,2);
	XtManageChild(main=XmCreateForm(Mshell,"main",NULL,0));

	menu = XmCreatePulldownMenu(Mshell,"menu",NULL,0);
	setupLogTypes(menu);
	XtSetArg(args[0],XmNsubMenuId,menu);
	XtManageChild(button=XmCreateOptionMenu(main,"viewMenu",args,1));

	XtManageChild(button=XmCreatePushButton(main,"saveButton",NULL,0));
	XtAddCallback(button,XmNactivateCallback,SaveLogCB,NULL);

	XtManageChild(work=XmCreateDrawingArea(main,"work",NULL,0));

	XtSetArg(args[0],XmNorientation,XmHORIZONTAL);
	XtSetArg(args[1],XmNshowValue,True);
	XtSetArg(args[2],XmNminimum,1);
	XtSetArg(args[3],XmNvalue,AverageWidth);
	XtManageChild(scale=XmCreateScale(main,"movingAvrg",args,4));
	XtAddCallback(scale,XmNvalueChangedCallback,ScaleCB,NULL);

	XtPopup(Mshell,XtGrabNone);
	MWin = XtWindow(work);
	XGetGeometry(display,MWin,&root,&x,&y,&Mwidth,&Mheight,&border,&depth);
	MPmap = XCreatePixmap(display,root,Mwidth,Mheight,depth);
	XtVaGetValues(work,XmNforeground,&Mfg,XmNbackground,&Mbg,NULL);
	XtVaGetValues(trendBtn,XmNfontList,&fontList,NULL);
	XmFontListInitFontContext(&context, fontList);
	XmFontListGetNextFont(context, &charset, &Gfont);
	MXoffset = XTextWidth(Gfont," 9.9e+99",8);
	MYspan = Mheight - MYoffset -
	 (Gfont->max_bounds.ascent+Gfont->max_bounds.descent)*2 - 6;
	MXtitleX = (Mwidth-MXoffset-XTextWidth(Gfont,MXtitle,strlen(MXtitle)))
	 /2 + MXoffset;
	MXtitleY = Mheight - Gfont->max_bounds.descent - 2;
	MXbase = MYoffset + MYspan + Gfont->max_bounds.ascent + 2;
	XSetForeground(display,gc,Mbg);
	XFillRectangle(display,MPmap,gc,0,0,Mwidth,Mheight);
	XtAddCallback(work,XmNexposeCallback,redisplayMonitorCB,NULL);
}
static void openTrendWindow(void) {
	if (!Mshell) createTrendShell();
	else XtPopup(Mshell,XtGrabNone);
}
static void closeTrendWindow(void) {
	XtPopdown(Mshell);
}
static void OpenTrendCB(w,cl,cd) Widget w; XtPointer cl, cd; {
	Boolean value;
	XtVaGetValues(w,XmNset,&value,NULL);
	if (value) reserveProc(openTrendWindow);
	else reserveProc(closeTrendWindow);
}
void setupMenuItemTrend(Widget menu) {
	XtManageChild(trendBtn=XmCreateToggleButton(menu,"Trend",NULL,0));
	XtAddCallback(trendBtn,XmNvalueChangedCallback,OpenTrendCB,NULL);
}
static int get_tic_span_int(int	max) {
	int	exp = 1;
	while (max >= 10) { max = (max + 9) / 10; exp *= 10; }
	if (exp < 10) return 1;
	else if (max < 2) return exp/10;
	else if (max < 3) return 2 * exp/10;
	else if (max < 5) return 5 * exp/10;
	else return exp;
}
static double get_tic_span_double(double min, double max, double *start) {
	double	span = max - min, exp = 1.;
	if (span <= 0) { *start = min; return 1.; }
	if (span > 1.) while (span >= 10.) { span /= 10.; exp *= 10.; }
	else while (span < 1.) { span *= 10.; exp /= 10.; }
	if (span < 2.) exp *= .1;
	else if (span < 3.) exp *= .2;
	else if (span < 5.) exp *= .5;
	*start = ceil(min / exp) * exp;
	return exp;
}
#define get_x(i) (((i) * (Mwidth-MXoffset)) / xspan + MXoffset)

static int get_y(double v, LogData *ld) {
	return (int)((ld->min==ld->max)?0.5:
		(1.-(v-ld->min)/(ld->max-ld->min))*MYspan)+MYoffset;
}
static void drawLogData(LogData *ld) {
	int	i, j, yorigin = MYoffset+MYspan, xspan = 10;
	int	itic, oldx, newx, oldy, newy, cmy, yy, cm[100];
	double	dtic, dstart;
	char	*fmt, buf[32];
	extern	unsigned long	Step;
	XSetForeground(display,gc,Mbg);
	XFillRectangle(display,MPmap,gc,0,0,Mwidth,Mheight);
	XSetForeground(display,gc,Mfg);
	XSetFont(display,gc,Gfont->fid);
	XDrawString(display,MPmap,gc,
		MXtitleX,MXtitleY,MXtitle,strlen(MXtitle));
	XDrawLine(display,MPmap,gc,MXoffset,yorigin,MXoffset,0);
	XDrawLine(display,MPmap,gc,MXoffset,yorigin,Mwidth-1,yorigin);
	if (Step > xspan) xspan = Step;
	itic = get_tic_span_int(xspan);
	for (i = 0; i <= Step; i += itic) {
		oldx = get_x(i);
		XDrawLine(display,MPmap,gc,oldx,yorigin-5,oldx,yorigin);
		sprintf(buf,"%d",i);
		oldx -= XTextWidth(Gfont,buf,strlen(buf))/2;
		XDrawString(display,MPmap,gc,oldx,MXbase,buf,strlen(buf));
	}
	dtic = get_tic_span_double(ld->min,ld->max,&dstart);
	if (dtic >= 1.) fmt = "%g";
	else if (dtic >= 0.1) fmt = "%.1f";
	else if (dtic >= 0.01) fmt = "%.2f";
	else fmt = "%.3f";
	for (; dstart <= ld->max; dstart += dtic) {
		oldy = get_y(dstart, ld);
		XDrawLine(display,MPmap,gc,MXoffset,oldy,MXoffset+5,oldy);
		sprintf(buf, fmt, dstart);
		oldx = MXoffset - XTextWidth(Gfont,buf,strlen(buf)) - 2;
		XDrawString(display,MPmap,gc,oldx,oldy,buf,strlen(buf));
	}
	for (j = 0; j < ld->n; j ++) {
		oldx = MXoffset;
		oldy = get_y(ld->data[j], ld);
		cm[0] = cmy = oldy;
		for (i = 1; i <= Step; i ++) {
			yy = get_y(ld->data[i*ld->n + j], ld);
			cmy += yy;
			if (i < AverageWidth) newy = cmy / (i + 1);
			else {
				cmy -= cm[i % AverageWidth];
				newy = cmy / AverageWidth;
			}
			newx = get_x(i);
			XDrawLine(display,MPmap,gc,oldx,oldy,newx,newy);
			cm[i % AverageWidth] = yy;
			oldy = newy;
			oldx = newx;
		}
	}
	XCopyArea(display,MPmap,MWin,gc,0,0,Mwidth,Mheight,0,0);
}
void showTrend(void) {
	extern unsigned long Step;
	Boolean value;
	XtVaGetValues(trendBtn,XmNset,&value,NULL);
	if (value) drawLogData(currentViewLog);
}
