#include <stdio.h>
#include "war.h"
#include "hash.h"

int InitialPopulation = 100;
double	InitialFoodStock = 200.;
#define FoodCosumingPerPerson	1.
#define FoodProductionPerLand	100.
static Village *freeV = NULL;
double FoodProductionPerWorker = 2.0;
double FoodSavingRate = 0.95;
double PopulationIncreasingRate = 1.1;
double FluctuationOfPopulation = 0.05;
double FluctuationOfProduction = 0.05;
int MaxSizeOfVillage = 50;
unsigned long villageID = 0;
unsigned long TracedVillageID = 0;

Village *newVillage(void) {
	Village *newV;
	if (!freeV) {
		int i;
		freeV = (Village *)malloc(sizeof(Village) * AllocUnit);
		if (!freeV) { perror("malloc for new village"); exit(0); }
		memset(freeV, 0, sizeof(Village) * AllocUnit);
		for (i = 0; i < AllocUnit-1; i ++)
			freeV[i].post = &freeV[i + 1];
		freeV[AllocUnit-1].post = NULL;
	}
	newV = freeV;
	freeV = freeV->post;
	newV->pre = newV->post = NULL;
	if (!newV->border) newV->border = newHash();
	newV->id = villageID ++;
	newV->ageM = 0;
	newV->l = NULL;
	newV->nLands = 0;
	newV->sv = InitialFoodStock;
	newV->p = InitialPopulation;
	setRandomMeme(newV->meme, sizeof(newV->meme));
	return newV;
}
static void deallocVillage(Village *v) {
	Land *l, *ll;
	for (l = v->l; l; l = ll) {
		ll = l->post;
		l->pre = l->post = NULL;
		l->v = NULL;
	}
	releaseHashItems(v->border);
}
releaseVillages(Village *v) {
	if (v) {
		Village *p, *last;
		for (p = v; p; p = p->post) {
			last = p;
			deallocVillage(p);
		}
		last->post = freeV;
		freeV = v;
	}
}
void calcSurplus(Village *v) {
	v->surplus = v->sv - v->p * FoodCosumingPerPerson;
}
#define	MutationInterval(v)	(v->meme[MemeVMutIntvl] / 10 + 3)
#define ThPlant(v)	(v->meme[MemeVPlant] * 2.2 / 255. + 0.8)
#define	ThMove(v)	(v->meme[MemeVMove] * 2.0 / 255. + 1.0)
#define	SaveFoodsR(v)	(v->meme[MemeVSaveFoods] * 0.8 / 255.)
void regularMutationVillage(Village *v) {
	if ((++ v->ageM) >= MutationInterval(v)) {
		mutatedCopy(v->meme, v->meme, sizeof(v->meme));
		v->ageM = 0;
	}
}
void deliveryFoodsForVillage(Village *v) {
	long	p;
	if (v->sv - v->fd < 0.) {
		printf("sv=%g, fd=%g, surplus=%g\n",
			v->sv,v->fd,v->surplus);
		exit(0);
	}
	v->sv -= v->fd;
	p = v->sv * (1. - SaveFoodsR(v)) / FoodCosumingPerPerson;
	if (p < v->p) {
		// killed by starvation
		v->p = p;
	}
}
static void gatherCountryInHashTable(void *c, void *ht) {
	Village *v = ((Land *)c)->v;
	if (v) {
		Country *c = v->k;
		putHash((long)c, (void *)c, ht);
	}
}
static void gatherCountryInArray(void *c, void *a) {
	Country ***cp = (Country ***)a;
	**cp = (Country *)c;
	(*cp) ++;
}
static int getNeighborCountries(Village *v, Country ***cl) {
	static void *ht = NULL;
	Country **pp, **nList;
	int	n;
	if (!ht) ht = newHash();
	putHash((long)v->k, (void *)v->k, ht);
	forAllHashItems(gatherCountryInHashTable, ht, v->border);
	n = countHashItems(ht);
	if (n <= 0) {
		fprintf(stderr,"countHashItems returned %d\n",n);
		exit(0);
	}
	*cl = pp = (Country **)allocTempMem(sizeof(void *) * n);
	forAllHashItems(gatherCountryInArray, &pp, ht);
	releaseHashItems(ht);
	return n;
}
void decideVillageAction(Village *v) {
	double	f1, f2, r;
	long	p, i, n;
	Country	**cl;
// initialize
	v->attacked = NoAttacked;
	if (v->p <= 0) return;
// consume foods
	v->sv -= FoodCosumingPerPerson * v->p;
	if (v->sv < 0.) v->sv = 0.;
// roten foods
	v->sv *= FoodSavingRate;
// produce foods
	f1 = FoodProductionPerWorker * v->pp;
	f2 = FoodProductionPerLand * v->nLands;
	v->sv += ((f1 < f2)? f1 : f2)
	 * (gaussianRandom() * FluctuationOfProduction + 1.);;
// decide plantation
	v->plantation = (f1 / f2 > ThPlant(v));
// decide country change
	n = getNeighborCountries(v, &cl);
	v->moveTo = NULL;
	r = v->richness = v->sv / v->p;
	for (i = 0; i < n; i ++) if (r < cl[i]->richness) {
		r = cl[i]->richness;
		v->moveTo = cl[i];
	}
	if (r / v->k->richness <= ThMove(v))
		v->moveTo = v->k;
}
extern void checkCBorderNew(void *, void *);
extern void checkCBorderSub(void *, void *);
extern void checkCBorderRVSub(void *, void *);
extern void checkNewVilBorder(Village *, Village *);
extern void checkSplitBorder(void *, void *);
struct boderAndVillage { void *border; Village *v; };
static void checkCBorderOld(void *item, void *udata) {
	Land *l = (Land *)item;
	struct boderAndVillage *bv = (struct boderAndVillage *)udata;
	if (l->v == bv->v) putHash((long)l, (void *)l, bv->border);
}
static void checkCBorder(Country *oldC, Country *newC, Village *v) {
	Village *u;
	struct countries { Country *oldC, *newC; } cc;
	struct boderAndVillage bv;
	forAllHashCells(checkCBorderNew, (void *)v, newC->border);
	cc.oldC = oldC; cc.newC = newC;
	forAllHashItems(checkCBorderSub, (void *)(&cc), v->border);
	bv.border = oldC->border; bv.v = v;
	for (u = oldC->v; u; u = u->post)
		forAllHashItems(checkCBorderOld, (void *)(&bv), u->border);
}
static void checkCBorderRV(Country *c, Village *v) {
	Village *u;
	struct boderAndVillage bv;
	struct { Country *c; Village *v; } cv;
	cv.c = c; cv.v = v;
	forAllHashItems(checkCBorderRVSub, (void *)(&cv), v->border);
	bv.border = c->border; bv.v = v;
	for (u = c->v; u; u = u->post)
		forAllHashItems(checkCBorderOld, (void *)(&bv), u->border);
}
static void removeVillage(Village *v) {
	if (v->pre) v->pre->post = v->post;
	else v->k->v = v->post;
	if (v->post) v->post->pre = v->pre;
}	
static void ruinVillage(Village *v) {
	Country *c = v->k;
	c->nVillages --;
	c->work = 1;
	checkCBorderRV(c, v);
	removeVillage(v);
	deallocVillage(v);
	v->post = freeV;
	freeV = v;
}
static void checkVillageSep(Land *land, Village *v1, Village *v2) {
	extern void traverseLand(Land *, Village *, int);
	Land	*l;
	int	i, k, n, mx, *nl;
	for (l = land; l; l = l->post) l->work = 0;
	for (l = land, n = 0; l; l = l->post)
		if (l->v == v1 && l->work == 0)
			traverseLand(l, v1, (++ n));
	if (n <= 1) return;
	nl = (int *)alloca(sizeof(int) * n);
	memset(nl, 0, sizeof(int) * n);
	// count number of land grids
	for (l = land; l; l = l->post)
		if (l->work > 0) nl[l->work - 1] ++;
	// find the broadest island
	for (mx = nl[0], k = 1, i = 1; i < n; i ++)
		if (mx < nl[i]) { mx = nl[i]; k = i + 1; }
	// change village of narrow islands
	for (l = land; l; l = l->post)
		if (l->work != k) l->v = v2;
}
static void splitVillage(Village *v) {
	Country	*c = v->k;
	Land	*l, *ll;
	Village	*nv;
	double sx = 0., sy = 0., vx = 0., vy = 0.;
	int	z, sz, nLandsOrig,
		whichNew = (lrand48() & 1) * 2 - 1;
	struct { Village *v, *nv; } vs;
	for (l = v->l; l; l = l->post) {
		sx += l->x; sy += l->y;
		vx += l->x * l->x; vy += l->y * l->y;
	}
	sx /= v->nLands; sy /= v->nLands;
	vx = vx / v->nLands - sx * sx;
	vy = vy / v->nLands - sy * sy;
	nv = newVillage();
	nv->k = c;
	mutatedCopy(nv->meme, v->meme, sizeof(v->meme));
	nv->l = NULL;
	for (l = v->l; l; l = l->post) {
		if (vx > vy) { z = l->x; sz = sx; }
		else { z = l->y; sz = sy; }
		if ((whichNew < 0)? (z < sz) : (z > sz))
			l->v = nv;
	}
	checkVillageSep(v->l, v, nv);
	checkVillageSep(v->l, nv, v);
	nLandsOrig = v->nLands;
	for (l = v->l; l; l = ll) {
		ll = l->post;
		if (l->v == nv) {
			v->nLands --; nv->nLands ++;
			if (l->pre) l->pre->post = l->post;
			else v->l = l->post;
			if (l->post) l->post->pre = l->pre;
			l->pre = NULL;
			l->post = nv->l;
			if (nv->l) nv->l->pre = l;
			nv->l = l;
		}
	}
	vs.v = v; vs.nv = nv;
	forAllHashCells(checkSplitBorder, (void *)(&vs), v->border);
	checkNewVilBorder(v, nv);
	checkNewVilBorder(nv, v);
	nv->p = nv->nLands * v->p / nLandsOrig;
	v->p -= nv->p;
	nv->sv = nv->nLands * v->sv / nLandsOrig;
	v->sv -= nv->sv;
	nv->pre = NULL;
	nv->post = c->v;
	c->v->pre = nv;
	c->v = nv;
	c->nVillages ++;
	setSimilarRGB(&nv->rgb, &v->rgb);
}
static void independence(Village *v) {
	Country *oldC = v->k, *newC = newCountry();
//	printf("Village%d gets independent from %d as %d.\n",
//		v->id, oldC->id, newC->id);
	removeVillage(v);
	v->post = v->pre = NULL;
	v->k = newC;
	newC->v = v;
	mutatedCopy(newC->meme, oldC->meme, sizeof(oldC->meme));
	registarCountry(newC);
	setSimilarRGB(&newC->rgb, &oldC->rgb);
	newC->nVillages ++;
	oldC->nVillages --;
	oldC->work = 1;
	checkCBorder(oldC, newC, v);
	oldC->mil -= v->ps; oldC->p -= v->p;
	newC->mil = v->ps; newC->p = v->p;
	oldC->nVindependent ++;
}
void moveCountry(Village *v, Country *oldC, Country *newC) {
	removeVillage(v);
	v->k = newC;
	v->pre = NULL;
	v->post = newC->v;
	if (newC->v) newC->v->pre = v;
	newC->v = v;
	newC->nVillages ++;
	oldC->nVillages --;
	oldC->work = 1;
	checkCBorder(oldC, newC, v);
	if (oldC->mil < v->ps || oldC->p < v->p)
		printf("moveCountry:C%d,v%d. vps=%d, mil=%d, vp=%d, cp=%d\n",
			oldC->id,v->id,v->ps,oldC->mil,v->p,oldC->p);
	oldC->mil -= v->ps; oldC->p -= v->p;
	newC->mil += v->ps; newC->p += v->p;
	oldC->nVmoveOut ++;
	newC->nVmoveIn ++;
}
void doVillageAction(Village *v) {
	if (v->p <= 0) {
		ruinVillage(v); return;
	}
	if (v->plantation) {
		Land *nl = (Land *)randomItemInHash(emptyLand, v->border);
		if (nl) {
			occupyLand(nl, v);
			if (v->nLands > MaxSizeOfVillage) splitVillage(v);
		}
	}
	if (v->attacked != AttackedThenOccupied) {
		if (!v->moveTo) {
			if (FunctionSwitch & FnIndependence) {
				if (v->pre || v->post) independence(v);
				else {
					unsigned char *m = v->k->meme;
					mutatedCopy(m, m, sizeof(v->k->meme));
				}
			}
		} else if ((FunctionSwitch & FnMoveCountry)
			 && v->k != v->moveTo)
			moveCountry(v, v->k, v->moveTo);
	}
	v->p = PopulationIncreasingRate * v->p
	 * (gaussianRandom() * FluctuationOfPopulation + 1.);
}
void showVillage(Village *v) {
	printf("Village #%d\n", v->id);
	printf("  Country: #%d, Size: %d,", v->k->id, v->nLands);
	printf(" Border cells: %d\n", countHashItems(v->border));
	printf("  Population: %d (o:%d, s:%d, p:%d)\n",
		v->p, v->po, v->ps, v->pp);
	printf("  Food stock: %g, Richness:%g\n", v->sv, v->richness);
	printf("  Plantation: %s\n", v->plantation? "yes" : "no");
	printf("  Move country: ");
	if (!v->moveTo) printf("independence\n");
	else if (v->moveTo != v->k) printf("#%d\n",v->moveTo->id);
	else printf("no\n");
}
